package tiimage;

/*
 * Copyright (c) 2014 Max Gttlicher
 *
 * This software is provided 'as-is', without any express or implied warranty. In no event will the authors be held liable for any damages arising from the use of this software.
 * 
 * Permission is granted to anyone to use this software for any purpose, including commercial applications, and to alter it and redistribute it freely, subject to the following restrictions:
 * 
 * The origin of this software must not be misrepresented; you must not claim that you wrote the original software. If you use this software in a product, an acknowledgment in the product documentation would be appreciated but is not required.
 *     Altered source versions must be plainly marked as such, and must not be misrepresented as being the original software.
 * This notice may not be removed or altered from any source distribution.
 */

import java.awt.Color;
import java.awt.image.BufferedImage;

/**
 * The class TIImage is a parser for TI.Image strings in both directions. It can
 * convert an image to a string which can be loaded as an image in TI-Nspire Lua
 * and such strings can be converted to images. Supports compressed strings
 * (\000\000 becomes \0\000) as far as supported by TI-Nspire.
 * 
 * @see <a
 *      href="http://wiki.inspired-lua.org/TI.Image">http://wiki.inspired-lua.org/TI.Image</a>
 *      for more Information on the TI.Image format
 */
public class TIImage {

	//I'm using char as it is unsigned. Also it's easier to create a string from chars
	private char[] data;
	
	private int w, h;

	/**
	 * Constructs a new TIImage from a TI.Image string. Removes surrounding
	 * quotes.
	 * 
	 * @param source
	 *            the TI.Image string containing the header and pixel data.
	 * @throws TIImageException
	 *             when the string can't be converted
	 */
	public TIImage(String source) throws TIImageException {
		// Remove surrounding "

		if (source.charAt(0) == '"' || source.charAt(0) == '\'') {
			source = source.substring(1);
		}

		if ((source.charAt(source.length() - 1) == '"' || source.charAt(source.length() - 1) == '\'')
				&& source.charAt(source.length() - 1) != '\\') {
			source = source.substring(0, source.length() - 1);
		}

		int width = 0, height = 0, pos = 0;

		// Read width and height to find out the length of the resulting array
		int size;
		try {
			for (int i = 0; i < 4; ++i) {
				width += ((readByte(source, pos)) << i * 8);
				pos = nextByte(source, pos);
			}
			for (int i = 0; i < 4; ++i) {
				height += ((readByte(source, pos)) << i * 8);
				pos = nextByte(source, pos);
			}

			w = width;
			h = height;
			
			System.out.printf("Loading image %dx%d", w, h);

			size = 20 + (width * height) * 2;
			data = new char[size];
		} catch (Exception e) {
			//Header can't be read
			e.printStackTrace();
			throw new TIImageException("Invalid header: "
					+ source.substring(0, pos));
		}
		
		if(w > 2000 || h > 2000) {
			throw new TIImageException("Too large dimensions");
		}
		
		pos = 0;
		int i = 0;

		//Read all the bytes of the string
		while (pos < source.length() && i < size) {
			data[i] = (char) readByte(source, pos);
			++i;
			pos = nextByte(source, pos);
		}

	}

	/**
	 * Constructs a new TI.Image from an existing image. The conversion can
	 * cause some loss in color and details.
	 * 
	 * @param source
	 *            the image to be converted
	 */
	public TIImage(BufferedImage source) {
		this(source, 128);
	}

	/**
	 * Constructs a new TI.Image from an existing image. The conversion can
	 * cause some loss in color and details.
	 * 
	 * @param source
	 *            the image to be converted
	 * @param alpha
	 *            pixels that are more transparent than alpha will not be drawn
	 */
	public TIImage(BufferedImage source, int alpha) {

		int width = source.getWidth();
		int height = source.getHeight();

		w = width;
		h = height;
		int line = 2 * width;

		data = new char[line * height + 20];

		// Writing the header:

		intInByte(width, data, 0); // 4 bytes representing the width of the
									// image

		intInByte(height, data, 4); // Next 4 bytes represent the height

		data[8] = 0; // Alignment
		data[9] = 0; // Flags
		data[10] = 0; // Padding
		data[11] = 0; // Padding

		intInByte(line, data, 12); // Length of a line in bytes

		shortInByte((short) 16, data, 16); //Bits per pixel

		shortInByte((short) 1, data, 18); //Planes per bit

		// Conversion of Pixel data

		// ColorModel model = source.getColorModel();

		int pos = 20;

		for (int i = 0; i < width * height; ++i) {
			Color c = new Color(source.getRGB(i % width, i / width), true);
			// int color = source.getRGB(i % width, i/width);
			int r = c.getRed();
			int g = c.getGreen();
			int b = c.getBlue();
			int a = c.getAlpha();

			short pxl = convertPixel(r, g, b, a, alpha);
			shortInByte(pxl, data, pos);

			pos += 2;

		}

		// for(int y=0;y<height;++y) {
		// for(int x=0;x<width;++x) {
		// Color c = new Color(source.getRGB(x, y), true);
		// System.out.print(c.getAlpha() < 128 ? "T" : 'O');
		// }
		// System.out.println();
		// }
	}

	/**
	 * Returns the width of the image.
	 * 
	 * @return the width of the image in pixels
	 */
	public int getWidth() {
		return w;
	}

	/**
	 * Returns the height of the image.
	 * 
	 * @return the height of the image in pixels
	 */
	public int getHeight() {
		return h;
	}

	/**
	 * Finds the end of an escape sequence
	 * 
	 * @param source
	 *            the string containing the sequence
	 * @param pos
	 *            the starting position of the sequence (first '\')
	 * @return the end of the sequence, i.e. the first non-digit or the fifth
	 *         character after the beginning of the sequence
	 */
	private int escapeEnd(String source, int pos) {
		int endEscape = pos + 2;

		while ((endEscape < (source.length()))
				&& Character.isDigit(source.charAt(endEscape))) {
			++endEscape;
		}
		;

		endEscape = Math.min(endEscape, pos + 4);
		return endEscape;
	}

	/**
	 * Reads one byte formatted in TI.Image.
	 * 
	 * @param source
	 *            the source string
	 * @param pos
	 *            the starting position of the byte in the string
	 * @return the value of the byte
	 * @throws TIImageException
	 *             when the byte cannot be read
	 */
	int readByte(String source, int pos) throws TIImageException {

		char c;
		try {
			c = source.charAt(pos);
		} catch (IndexOutOfBoundsException e1) {
			e1.printStackTrace();
			throw new TIImageException("wrong size");
		}
		if (c == '\\') {
			if (source.charAt(pos + 1) == '\\') {
				return '\\';
			}
			if (source.charAt(pos + 1) == '"') {
				return '"';
			}
			if (source.charAt(pos + 1) == '\'') {
				return '\'';
			}


			try {
				return Integer.parseInt(source.substring(pos + 1,
						escapeEnd(source, pos)));
			} catch (NumberFormatException e) {
				e.printStackTrace();
				throw new TIImageException("Cannot parse \""
						+ source.substring(pos + 1, escapeEnd(source, pos))
						+ "\" at position " + pos);
			}
		} else {
			return (byte) c;
		}
	}

	/**
	 * Returns the starting index of the succeeding byte.
	 * 
	 * @param source
	 *            the source string
	 * @param pos
	 *            the string position of the byte in the string
	 * @return the starting position of the next byte in the string
	 */
	int nextByte(String source, int pos) {
		char c = source.charAt(pos);
		if (c == '\\') {
			if (source.charAt(pos + 1) == '\\') {
				return pos + 2;
			}
			if (source.charAt(pos + 1) == '"') {
				return pos + 2;
			}
			if (source.charAt(pos + 1) == '\'') {
				return pos + 2;
			}

			return escapeEnd(source, pos);

		} else {
			return pos + 1;
		}
	}

	/**
	 * Converts an RGBA-Color value to 15-bit high color with alpha bit.
	 * 
	 * @param r
	 *            the original red value
	 * @param g
	 *            the original green value
	 * @param b
	 *            the original blue value
	 * @param a
	 *            the original transparency
	 * @param malpha
	 *            the pixel will be rendered if alpha is higher then this
	 * @return the converted color in 2 bytes (1 short)
	 */
	private short convertPixel(int r, int g, int b, int a, int malpha) {
		short r5 = (short) ((r / 8));
		short g5 = (short) ((g / 8));
		short b5 = (short) ((b / 8));
		short a1 = (short) ((a < malpha) ? 0 : 1);

		short result = 0;
		result += b5;
		result += (g5 << 5);
		result += (r5 << 10);
		result += (a1 << 15);

		return result;
	}

	/**
	 * Writes an integer to an array of char in LittleEndian byte order
	 * 
	 * @param from
	 *            the integer to be converted to LittleEndian
	 * @param to
	 *            the receiving array
	 * @param pos
	 *            the position to be written to
	 */
	void intInByte(int from, char[] to, int pos) {
		to[pos + 0] = (char) (from);
		to[pos + 1] = (char) (from >> 8);
		to[pos + 2] = (char) (from >> 16);
		to[pos + 3] = (char) (from >> 24);
	}

	/**
	 * Writes a short to an array of char in LileEndian byte order.
	 * 
	 * @param from
	 *            the short to be converted
	 * @param to
	 *            the receiving arrray
	 * @param pos
	 *            the position to be written to
	 */
	void shortInByte(short from, char[] to, int pos) {
		to[pos + 0] = (char) (0x00FF & from);
		to[pos + 1] = (char) ((0xFF00 & from) >> 8);
	}

	/**
	 * Checks whether a character is printable or not.
	 * 
	 * @param c
	 *            the character to be checked
	 * @return true if the character is printable.
	 */
	private boolean isPrintable(char c) {
		return (c > 0x20 && c < 0x7E);
	}

	/**
	 * Converts a byte to a string in TI.Image format
	 * 
	 * @param c
	 *            the byte to be converted
	 * @return a string supported by TI.Image
	 */
	public String byteToString(int c) {
		return byteToString(c, -1);
	}

	/**
	 * Converts a byte to a string in TI.Image format with byte compression
	 * 
	 * @param c
	 *            the byte to be converted
	 * @param next
	 *            the succeeding byte, used to determine whether the escape
	 *            sequence can be compressed or not
	 * @return a string supported by TI.Image
	 */
	public String byteToString(int c, int next) {
		String bt;
		if (!isPrintable((char) c)) {
			if (!isPrintable((char) next)) {
				bt = String.format("\\%d", c);
			} else {
				bt = String.format("\\%03d", c);
			}
		} else {
			if (c == 0x22) {
				bt = "\\\"";
			} else if (c == 0x5C) {
				bt = "\\\\";
			} else if (c == 0x27) {
				bt = "\\\'";
			} else {
				bt = String.format("%c", c);
			}
		}
		return bt;
	}

	/**
	 * Returns the color stored in 2 bytes in LittleEndian order
	 * 
	 * @param byte1
	 *            the first byte
	 * @param byte2
	 *            the second byte
	 * @return the RGBA-Color
	 */
	public Color getColor(int byte1, int byte2) {
		int d = 0;
		d += byte2 << 8;
		d += byte1;
		
		int b = ((d) % 32) * 8;
		int g = ((d >> 5) % 32) * 8;
		int r = ((d >> 10) % 32) * 8;
		int a = ((d >> 15) % 2) == 1 ? 0xFF : 0x00;
		
		return new Color(r, g, b, a);
	}

	/**
	 * Creates a drawable image from the data stored. Expect some loss of
	 * details and of depth of color compared to the original image.
	 * 
	 * @return a drawable image.
	 */
	public BufferedImage getImage() {
		BufferedImage img;
		img = new BufferedImage(w, h, BufferedImage.TYPE_INT_ARGB);
		
		int i = 20;
		for (int y = 0; y < h; ++y) {
			for (int x = 0; x < w; ++x) {
				Color c;
				c = getColor(data[i], data[i + 1]);
				img.setRGB(x, y, c.getRGB());
				i += 2;
			}
		}

		System.out.println("Image loaded successfully!");

		return img;
	}

	/**
	 * Returns a string representing the image. The string can be loaded into
	 * TI-Nspire Lua programs.
	 */
	public String getString() {
		String image = "";

		for (int i = 0; i < data.length; ++i) {
			int c = data[i];
			String bt = null;
			if (i < data.length - 1) {
				bt = byteToString(c, data[i + 1]);
			} else {
				bt = byteToString(c);
			}
			image = String.format("%s%s", image, bt);
		}
		return image;
	}

	/**
	 * Returns a string representing the object.
	 */
	public String toString() {
		return String.format("%s[width=%d,height=%d,data=\"%s\"]", this
				.getClass().toString(), w, h, getString());
	}

}
